// Exercise 2 -  A distributed pipelined consumer/producer problem
//          (consumers are also producers for next stage
//
// Use in Linux:
//     gcc 09a_sem_prod_cons.c -lpthread -lrt
//     ./a.out

#define NO_ITEMS_PRODUCED 1000
#define BUFF_SIZE 5

#define MAX_RAND 1000

#include <pthread.h>
#include <semaphore.h>
#include <stdio.h>
#include <stdint.h>
#include <sched.h>
#include <stdlib.h>

/* #define SHARED 1  - shared between processes */ 
#define SHARED 0   /* for Linux - shared between threads but not between processes */

/* the two threads */
void *Producer(void *);
void *Consumer(void *);

int buff [BUFF_SIZE];       /* shared buffer */

/* global semaphores  & locks */
sem_t empty, full;

// who has access to these? need locks?
long produced = 0;
long consumed = 0;

pthread_mutex_t L_prod = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t L_cons = PTHREAD_MUTEX_INITIALIZER;

/* main() -- read command line and create threads, then
             print result when the threads have quit */
int main(int argc, char *argv[]) {

	/* thread ids and attributes */
	pthread_t pid0, pid1, cid0, cid1;

	pthread_attr_t attr;
	pthread_attr_init(&attr);
	pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);

	sem_init(&empty, SHARED, BUFF_SIZE); // producer1/consumer1 semaphore (stops after 5 elements)
	sem_init(&full, SHARED, 0);  // producer1/consumer1 semaphore (initially stopped)
	

	pthread_create(&pid0, &attr, Producer, (void*) 0);
	pthread_create(&pid1, &attr, Producer, (void*) 1);
	pthread_create(&cid0, &attr, Consumer, (void*) 0);
	pthread_create(&cid1, &attr, Consumer, (void*) 1);

	pthread_join(pid0, NULL);
	pthread_join(pid1, NULL);
	pthread_join(cid0, NULL);
	pthread_join(cid1, NULL);

	sem_destroy(&empty);
	sem_destroy(&full);


	sem_close(&empty);
	sem_close(&full);

}

void *Producer(void *arg) { // deposit elements to buff
	int i = (intptr_t) arg;
	int element;
	int count = 0;

	if (i == 1) {
        	printf("                                  PROD_%d starts \n", i);
        	fflush(stdout);
	} else {
		printf("PROD_%d starts \n", i);
        	fflush(stdout);
	}

	//srand(1U);
	while (count < NO_ITEMS_PRODUCED) {
		sem_wait(&empty);
		pthread_mutex_lock(&L_prod);
			count++;
			element = rand() % MAX_RAND;
			buff [produced] = element;
			if (i == 1) {
                        	printf("                                  PROD_%d->buff[%d]:%ld (prod:%d) \n", i, produced, element, count);
        			fflush(stdout);
			} else {
                        	printf("PROD_%d->buff[%d]:%ld (prod:%d) \n", i, produced, element, count);
				fflush(stdout);
			}
			produced = (produced + 1) % BUFF_SIZE;
		pthread_mutex_unlock(&L_prod);
		sem_post(&full);

		if(lrand48()%2 == 1)  
			sched_yield(); /* random yield to another thread */ 
		//sleep(0.001);
	}
}

void *Consumer(void *arg) { // remove elements from buff (full? empty?)
        int i = (intptr_t) arg, x;
	int element;
	int count = 0;

	if (i == 1) {
        	printf("                                                                                                      CONS_%d starts \n", i);
        	fflush(stdout);
	} else {
		printf("                                                                    CONS_%d starts \n", i);
        	fflush(stdout);
	}

        while ((count < NO_ITEMS_PRODUCED)) {
                sem_wait(&full);
		pthread_mutex_lock(&L_cons);
			count++;
			element = buff[consumed];
			if (i == 1) {
                        	printf("                                                                                                      CONS_%d<-buff[%d]:%ld (cons:%d) \n", i, consumed, element, count);
        			fflush(stdout);
			}
			else {
                        	printf("                                                                    CONS_%d<-buff[%d]:%ld (cons:%d) \n", i, consumed, element, count);
                        	fflush(stdout);
			}
			consumed = (consumed + 1) % BUFF_SIZE;
		pthread_mutex_unlock(&L_cons);
		sem_post(&empty);

		//sleep(0.001);

                if(lrand48()%2 == 1) 
                        sched_yield(); /* random yield to another thread */

        }
}
